#include "StdAfx.h"
#include<GL/gl.h>
#include<GL/glu.h>
#include <algorithm>
#include "moebiusstrip.h"


MoebiusStrip::MoebiusStrip(void)
{
 BlockNum=12;
 SideNum=2;
 Turns=1;
 Radius=4;
 InnerRadius=.8f;
 FilletRatio=.1f;
 StartTwistDeg=0;
 StepNum=4;
}

MoebiusStrip::~MoebiusStrip(void)
{
}

void MoebiusStrip::XMLWrite(FILE *fp)
{
 	fprintf(fp,"<MoebiusStrip\n");
	fprintf(fp,"  SideNum       = \"%i\"\n",SideNum       );
	fprintf(fp,"  Turns         = \"%f\"\n",Turns         );
	fprintf(fp,"  Radius        = \"%f\"\n",Radius        );
	fprintf(fp,"  BlockNum      = \"%i\"\n",BlockNum      );
	fprintf(fp,"  InnerRadius   = \"%f\"\n",InnerRadius   );
	fprintf(fp,"  FilletRatio   = \"%f\"\n",FilletRatio   );
	fprintf(fp,"  StartTwistDeg = \"%f\"\n",StartTwistDeg );
	fprintf(fp,"  StepNum       = \"%i\"\n",StepNum       );
	fprintf(fp,"/>\n");
}

void MoebiusStrip::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "MoebiusStrip");
	SideNum       =        atoi(xml["SideNum"].c_str());
	Turns         = (float)atof(xml["Turns"].c_str());
	Radius        = (float)atof(xml["Radius"].c_str());
	BlockNum      =        atoi(xml["BlockNum"].c_str());
	InnerRadius   = (float)atof(xml["InnerRadius"].c_str());
	FilletRatio   = (float)atof(xml["FilletRatio"].c_str());
	StartTwistDeg = (float)atof(xml["StartTwistDeg"].c_str());
	StepNum       =        atoi(xml["StepNum"].c_str());
}

void MoebiusStrip::Clear()
{
	Ring.resize(BlockNum*SideNum);
	for(int i=0;i<SideNum;++i)
		for(int j=0;j<BlockNum;++j)
			Sec(i,j).Clear();
}

void MoebiusStrip::glDraw(const float DocTime)
{
	int nm=0;
	vector<CIMesh>::iterator mi;
	for(mi=Ring.begin();mi!=Ring.end();++mi)
	{
		glLoadName(nm++);
		glColor((*mi).c);
		(*mi).Draw<true,true>();
	}
	glLoadName(-1);
	
}	
	

void MoebiusStrip::Generate()
{
	vector<Point3f> P;
	vector<Point3f> N;
	
	Clear();
	GenerateSideSection(P,N);
	for(int i=0;i<SideNum;++i)
		for(int j=0;j<BlockNum;++j)
			GenerateBlock(P,N,i,j,Sec(i,j));
}
//
//
//
//
// Genera un dato blocchetto
//
//
void MoebiusStrip::GenerateBlock(const vector<Point3f> &SP, const vector<Point3f> &SN, int SideInd, int BlockInd, CIMesh& m)
{
	float BlockAngleDeg=360.0f/BlockNum;
	float StepAngleDeg=BlockAngleDeg/StepNum;
	float SideAngleDeg=(360.0f/SideNum);
  float StepTwistDeg=float(SideAngleDeg*Turns)/(BlockNum*StepNum);
	float BlockTwistDeg=float(SideAngleDeg*Turns)/BlockNum;
	
	// -- Disegno dell'Anello -- 
  // il disegno viene fatto a sezioni, 

	for(int i=0;i<StepNum+1;++i)
	{
	  Matrix44f ra; ra.Rotate(BlockAngleDeg*BlockInd+i*StepAngleDeg    , Point3f(0,0,1));// Rotazione Intorno all'asse principale
		Matrix44f ta; ta.Translate(Radius,0,0);
		Matrix44f rs; rs.Rotate(SideAngleDeg*SideInd +  BlockTwistDeg*BlockInd + StepTwistDeg*i, Point3f(0,1,0));// Avvolgimento della striscia.
		
		Matrix44f mm=ra*ta*rs;
			
		for(unsigned int j=0;j<SP.size();++j)
		{
			Point4f p4=mm*Point4f(SP[j][0],SP[j][1],SP[j][2],1); 
			Point4f n4=mm*Point4f(SN[j][0],SN[j][1],SN[j][2],0);
			m.vert.push_back(CIVertex(
				Point3f(float(p4[0]),float(p4[1]),float(p4[2])),
				Point3f(float(n4[0]),float(n4[1]),float(n4[2])),float(i),float(j)));
		}
	}
  //  a--c
	//  |\ |
	//  | \|
	//  b--d  
	//
	unsigned int sz=SP.size();
	for(int i=0;i<StepNum;++i)
	{
		for(unsigned int k=0;k<sz;++k)
		{
			CIFace tmp;

			tmp.v[0]=&(m.vert[(i+0)*sz+(k+0)%sz]); // a 
			tmp.v[1]=&(m.vert[(i+1)*sz+(k+1)%sz]); // d 
			tmp.v[2]=&(m.vert[(i+0)*sz+(k+1)%sz]); // b 
			m.face.push_back(tmp);

			tmp.v[0]=&(m.vert[(i+0)*sz+(k+0)%sz]); // a
			tmp.v[1]=&(m.vert[(i+1)*sz+(k+0)%sz]); // c
			tmp.v[2]=&(m.vert[(i+1)*sz+(k+1)%sz]); // d
			m.face.push_back(tmp);
		}
	}

}

void MoebiusStrip::Move(MoebiusStrip::Pos &p, float DistDeg)
{
	p.ThetaDeg+=DistDeg;
	if(p.ThetaDeg >360) 
	{
		p.ThetaDeg -= 360.0;
		p.CurSide += int(Turns);
	}
	if(p.ThetaDeg <0) 
	{
		p.ThetaDeg += 360.0;
		p.CurSide -= int(Turns);
	}
}

void MoebiusStrip::GenerateSideSection( vector<Point3f> &SP,  vector<Point3f> &SN)
{
	assert(SideNum>1);
	if(FilletRatio<0) FilletRatio=0;
	if(FilletRatio>1) FilletRatio=1;

	float TurnStepDeg=360.0f/SideNum;
	float StripAngleDeg=TurnStepDeg*(1.0f-FilletRatio);
	float HalfFilletAngleDeg=(TurnStepDeg*FilletRatio)/2;
	SP.clear();SN.clear();
	float BaseAngleDeg = StartTwistDeg-TurnStepDeg/2.0f;

	float a0Rad=ToRad(BaseAngleDeg+(0)*TurnStepDeg  );
	float a1Rad=ToRad(BaseAngleDeg+(0)*TurnStepDeg + HalfFilletAngleDeg);
	float a2Rad=ToRad(BaseAngleDeg+(1)*TurnStepDeg - HalfFilletAngleDeg);
	float a3Rad=ToRad(BaseAngleDeg+(1)*TurnStepDeg );
	float shrt=Cos(ToRad(HalfFilletAngleDeg));

	float n0Rad=ToRad(BaseAngleDeg+(0)*TurnStepDeg );
	float n1Rad=ToRad(BaseAngleDeg+(0)*TurnStepDeg + TurnStepDeg/2.0f);
	float n2Rad=ToRad(BaseAngleDeg+(1)*TurnStepDeg );

	SP.push_back(Point3f(Cos(a0Rad),0,Sin(a0Rad))*InnerRadius*shrt); SN.push_back(Point3f(Cos(n0Rad),0,Sin(n0Rad)));
	SP.push_back(Point3f(Cos(a1Rad),0,Sin(a1Rad))*InnerRadius); SN.push_back(Point3f(Cos(n0Rad),0,Sin(n0Rad)));
	SP.push_back(Point3f(Cos(a1Rad),0,Sin(a1Rad))*InnerRadius); SN.push_back(Point3f(Cos(n1Rad),0,Sin(n1Rad)));
	SP.push_back(Point3f(Cos(a2Rad),0,Sin(a2Rad))*InnerRadius); SN.push_back(Point3f(Cos(n1Rad),0,Sin(n1Rad)));
	SP.push_back(Point3f(Cos(a2Rad),0,Sin(a2Rad))*InnerRadius); SN.push_back(Point3f(Cos(n2Rad),0,Sin(n2Rad)));
	SP.push_back(Point3f(Cos(a3Rad),0,Sin(a3Rad))*InnerRadius*shrt); SN.push_back(Point3f(Cos(n2Rad),0,Sin(n2Rad)));
}

void MoebiusStrip::Transform(const MoebiusStrip::Pos &p)
{
	// di quanto si gira per ogni turn (e.g. 
	float TurnStepDeg = (360.0f/SideNum);
	// di quanto si avvolge per ogni degree
	float TwistPerDeg = (TurnStepDeg * Turns )/360.0f;
	float HalfFilletAngleDeg=(TurnStepDeg*FilletRatio)/2;
 	float BaseAngleDeg = StartTwistDeg-TurnStepDeg/2.0f;
	glRotatef(p.ThetaDeg, 0,0,1);   // Rotazione Intorno all'asse principale
	glTranslatef(Radius,0,0);
	glRotatef(TurnStepDeg*p.CurSide+p.ThetaDeg*TwistPerDeg, 0,1,0);   // Avvolgimento della striscia.
	glTranslatef(Cos(ToRad(BaseAngleDeg+HalfFilletAngleDeg))*InnerRadius,0,0);
}
///////////////////
CSGNode *CSGMb::Allocate(const string &classname)
{
	if(classname=="MoebiusStrip") {
		assert(m==0);
		m=new MoebiusStrip;
		return m;
	}
	if(classname=="CSGSphereMesh") 	return new CSGSphereMesh;
	if(classname=="CSGAnimMoebiusTransf") 	return new CSGAnimMoebiusTransf;
	CSGNode *nn=CSG::Allocate(classname);
  if(nn) return nn;
	return 0;
}

void CSGMb::Clear()
{
	CSG::Clear();
  m=0;
}

/////////////////////////////////////////
void CSGSphereMesh:: XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGSphereMesh\n");
	fprintf(fp,"  Radius = \"%f\" Parallel = \"%i\" Meridian= \"%i\"\n",
		Radius,Parallel,Meridian);
	fprintf(fp,"/>\n");
}

void CSGSphereMesh::glDraw(const float DocTime){
	Draw<true,true>();
}

void CSGSphereMesh::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "CSGSphereMesh");
	Radius  = (float)atof(xml["Radius"].c_str());
	Parallel =  atoi(xml["Parallel"].c_str());
	Meridian   = atoi(xml["Meridian"].c_str());
	CIShape::Sphere(*this,Radius,Parallel,Meridian);
}


void CSGAnimMoebiusTransf::glDraw(const float DocTime)
{
	MoebiusStrip::Pos pp;
	float CurAngleDeg =StartAngleDeg + DocTime*AngularSpeedDPS;
	pp.CurSide=StartSide;
	pp.ThetaDeg=CurAngleDeg;
	b->m->Transform(pp);
}

void CSGAnimMoebiusTransf:: XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGAnimMoebiusTransf\n");
	fprintf(fp,"  AngularSpeedDPS = \"%f\" StartAngleDeg = \"%f\" StartSide = \"%f\"\n",
		AngularSpeedDPS,StartAngleDeg,StartSide);
	fprintf(fp,"/>\n");
}

void CSGAnimMoebiusTransf::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "CSGAnimMoebiusTransf");
	AngularSpeedDPS = (float)atof(xml["AngularSpeedDPS"].c_str());
	StartAngleDeg   = (float)atof(xml["StartAngleDeg"].c_str());
	StartSide       =  atoi(xml["StartSide"].c_str());
	b=(CSGMb *)Base;
}
