#include "StdAfx.h"
#include "gl/gl.h"
#include "gl/glu.h"
#include <algorithm>
#include <vcg/Matrix44.h>
#include <vcg/xml/XMl.h>
//#include <vcg/xml/XMlutility.h>
#include "sg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

using namespace vcg;

CSG::CSG(void)
{
	LightCounter=0;
}

CSG::~CSG(void)
{
}


int CSG::Pick(int x, int y, const float &DocTime)
{
	long hits;	
  static unsigned int selectBuf[16384];
  glSelectBuffer(16384, selectBuf);
  glRenderMode(GL_SELECT);
  glInitNames();

  /* Because LoadName() won't work with no names on the stack */
  glPushName(-1);
	double mp[16];
	
  int viewport[4];
  glGetIntegerv(GL_VIEWPORT,viewport);
	glMatrixMode(GL_PROJECTION);
	glGetDoublev(GL_PROJECTION_MATRIX ,mp);
	glPushMatrix();
  glLoadIdentity();
  gluPickMatrix(x, viewport[3]-y, 4, 4, viewport);
  //gluPickMatrix(x, y, 4, 4, viewport);
	glMultMatrixd(mp);

	glMatrixMode(GL_MODELVIEW);

	root.glDraw(DocTime);
 
  glMatrixMode(GL_PROJECTION);
	glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  hits = glRenderMode(GL_RENDER);
  if (hits <= 0)     return -1;
  
	vector< pair<double,unsigned int> > H;
	for(int ii=0;ii<hits;ii++){
		TRACE("%ui %ui %ui %ui\n",selectBuf[ii*4],selectBuf[ii*4+1],selectBuf[ii*4+2],selectBuf[ii*4+3]);
		H.push_back( make_pair(selectBuf[ii*4+1]/4294967295.0,selectBuf[ii*4+3]));
	}
	sort(H.begin(),H.end());
  TRACE("\n Closest is %i\n",H[0].second);
  return H[0].second;
}


CSGNode *CSG::Allocate(const string &classname)
{
	CSGNode *pt=0;
	if(classname=="CSGLight") pt= new CSGLight(LightCounter++);
  if(classname=="CSGGroup") pt= new CSGGroup;
  if(classname=="CSGRotation") pt= new CSGRotation;
  if(classname=="CSGTranslation") pt= new CSGTranslation;
  if(classname=="CSGTransformation") pt= new CSGTransformation;
  if(classname=="CSGAnimRotation") pt= new CSGAnimRotation;
  if(classname=="CSGAnimZPrecession") pt= new CSGAnimZPrecession;
	return pt;
}

void CSG::Clear()
{
	root.Clear();
	LightCounter=0;
}

///////////////////// CSGGroup ////////////////////////
CSGGroup::~CSGGroup()
{
  Clear();
}

void CSGGroup::Clear()
{
 iterator i;
	for(i=Sons.begin();i!=Sons.end();++i)
	{
		delete (*i);
	}
}
void CSGGroup::glDraw(const float DocTime)
{
	glPushMatrix();
	if(SaveLightState) glPushAttrib(GL_LIGHTING_BIT);
  iterator i;
	for(i=Sons.begin();i!=Sons.end();++i)
		(*i)->glDraw(DocTime);
	if(SaveLightState) glPopAttrib();
	glPopMatrix();
}
void CSGGroup::XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGGroup SaveLightState = \"%i\" >\n",SaveLightState?1:0);
	iterator i;
	for(i=Sons.begin();i!=Sons.end();++i)
		(*i)->XMLWrite(fp);
	fprintf(fp,"</CSGGroup>\n");
}
void CSGGroup::XMLRead(Xml &xml, CSG *Base)
{
	Sons.clear();
 	assert(xml.id == "CSGGroup");
	if(xml["SaveLightState"]=="0") SaveLightState=false;
	Xml::iterator xi;
	for(xi = xml.begin(); xi != xml.end(); xi++) 
	{
		CSGNode *np=Base->Allocate((*xi).id);
		np->XMLRead(*xi,Base);
		Sons.push_back(np);
	}
}
///////////////////// CSGLight ////////////////////////

CSGLight::CSGLight(const int id)
{
	Id=id;
	Position      = Point4f(0,0,1,0);
	DiffuseColor  = Point4f(1,1,1,1);
	AmbientColor  = Point4f(0,0,0,1);
	SpecularColor = Point4f(1,1,1,1);
}

void  CSGLight::glDraw(const float DocTime)
{
	glEnable(GL_LIGHT0+Id);
	glLightfv(GL_LIGHT0+Id, GL_DIFFUSE,  DiffuseColor.V());
	glLightfv(GL_LIGHT0+Id, GL_AMBIENT,  AmbientColor.V());
	glLightfv(GL_LIGHT0+Id, GL_SPECULAR, SpecularColor.V());
	glLightfv(GL_LIGHT0+Id, GL_POSITION, Position.V());
}

void CSGLight::XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGLight\n");
	fprintf(fp,"DC0= \"%f\" DC1= \"%f\" DC2= \"%f\" DC3= \"%f\" \n",DiffuseColor[0] ,DiffuseColor[1],DiffuseColor[2],DiffuseColor[3]);
	fprintf(fp,"AC0= \"%f\" AC1= \"%f\" AC2= \"%f\" AC3= \"%f\" \n",AmbientColor[0] ,AmbientColor[1],AmbientColor[2],AmbientColor[3]);
	fprintf(fp,"SC0= \"%f\" SC1= \"%f\" SC2= \"%f\" SC3= \"%f\" \n",SpecularColor[0],SpecularColor[1],SpecularColor[2],SpecularColor[3]);
	fprintf(fp,"PT0= \"%f\" PT1= \"%f\" PT2= \"%f\" PT3= \"%f\" \n",Position[0],     Position[1],Position[2],Position[3]);
	fprintf(fp,"/>\n");
}
void CSGLight::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "CSGLight");
	DiffuseColor[0]  = (float)atof(xml["DC0"].c_str()); DiffuseColor[1]  = (float)atof(xml["DC1"].c_str());
	AmbientColor[0]  = (float)atof(xml["AC0"].c_str()); AmbientColor[1]  = (float)atof(xml["AC1"].c_str());
	SpecularColor[0] = (float)atof(xml["SC0"].c_str()); SpecularColor[1] = (float)atof(xml["SC1"].c_str());
	Position[0]      = (float)atof(xml["PT0"].c_str()); Position[1]      = (float)atof(xml["PT1"].c_str());

  DiffuseColor[2]  = (float)atof(xml["DC2"].c_str());DiffuseColor[3]  = (float)atof(xml["DC3"].c_str());
  AmbientColor[2]  = (float)atof(xml["AC2"].c_str());AmbientColor[3]  = (float)atof(xml["AC3"].c_str());
  SpecularColor[2] = (float)atof(xml["SC2"].c_str());SpecularColor[3] = (float)atof(xml["SC3"].c_str());
  Position[2]      = (float)atof(xml["PT2"].c_str());Position[3]      = (float)atof(xml["PT3"].c_str());
}

///////////////////// CSGTransformation ////////////////////////

void CSGTransformation::glDraw(const float DocTime){ glMultMatrix(m); };
void CSGTransformation::XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGTransformation>\n");
	::XMLWrite(fp,"M",m);
	fprintf(fp,"</CSGTransformation>\n");
}
void CSGTransformation::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "CSGTransformation");
	Xml::iterator xi;
	for(xi = xml.begin(); xi != xml.end(); xi++) 
	{
		if((*xi).id=="Matrix") ::XMLRead(*xi,"M",m);
	}
}
///////////////////// CSGAnimRotation ////////////////////////

void CSGAnimRotation ::glDraw(const float DocTime) 
{
	float CurAngleDeg= StartAngleDeg + DocTime*AngularSpeedDPS;
	glRotatef( CurAngleDeg, Axis[0],Axis[1],Axis[2]);  
};

void CSGAnimRotation::XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGAnimRotation\n");
	fprintf(fp,"  AxisX = \"%f\" AxisY = \"%f\" AxisZ = \"%f\" AngularSpeedDPS = \"%f\" StartAngleDeg= \"%f\"\n",
		Axis[0],Axis[1],Axis[2],AngularSpeedDPS,StartAngleDeg);
	fprintf(fp,"/>\n");
}
void CSGAnimRotation::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "CSGAnimRotation");
	Axis[0]         = (float)atof(xml["AxisX"].c_str());
	Axis[1]         = (float)atof(xml["AxisY"].c_str());
	Axis[2]         = (float)atof(xml["AxisZ"].c_str());
	AngularSpeedDPS = (float)atof(xml["AngularSpeedDPS"].c_str());
	StartAngleDeg   = (float)atof(xml["StartAngleDeg"].c_str());
}

///////////////////// CSGTranslation ////////////////////////


void CSGTranslation::XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGTranslation\n");
	fprintf(fp,"  VX = \"%f\" VY = \"%f\" VZ = \"%f\" \n", V[0], V[1], V[2]);
	fprintf(fp,"/>\n");
}

void CSGTranslation::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "CSGTranslation");
	V[0]    = (float)atof(xml["VX"].c_str());
	V[1]    = (float)atof(xml["VY"].c_str());
	V[2]    = (float)atof(xml["VZ"].c_str());
}
///////////////////// CSGRotation ////////////////////////

void CSGRotation::XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGRotation\n");
	fprintf(fp,"  AxisX = \"%f\" AxisY = \"%f\" AxisZ = \"%f\" AngleDeg= \"%f\"\n",
		Axis[0],Axis[1],Axis[2],AngleDeg);
	fprintf(fp,"/>\n");
}

void CSGRotation::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "CSGRotation");
	Axis[0]    = (float)atof(xml["AxisX"].c_str());
	Axis[1]    = (float)atof(xml["AxisY"].c_str());
	Axis[2]    = (float)atof(xml["AxisZ"].c_str());
	AngleDeg   = (float)atof(xml["AngleDeg"].c_str());
}
///////////////////// CSGAnimZPrecession ////////////////////////

CSGAnimZPrecession::CSGAnimZPrecession()
{
	StartAngleDeg=0;
	AngularSpeedDPS=10;
	DeclinationDeg=30;
}
 	
void CSGAnimZPrecession::glDraw(const float DocTime) 
{
	// moto di precessione: una rotazione il cui asse ruota intorno all'asse z
	float CurAngleRad=ToRad(StartAngleDeg + DocTime*AngularSpeedDPS);
	glRotated( DeclinationDeg,cos(CurAngleRad),sin(CurAngleRad),0);  
};
	
void CSGAnimZPrecession::XMLWrite(FILE *fp)
{
	fprintf(fp,"<CSGAnimZPrecession\n");
	fprintf(fp,"  DeclinationDeg = \"%f\" AngularSpeedDPS = \"%f\" StartAngleDeg= \"%f\"\n",
		DeclinationDeg,AngularSpeedDPS,StartAngleDeg);
	fprintf(fp,"/>\n");
}

void CSGAnimZPrecession::XMLRead(Xml &xml, CSG *Base)
{
 	assert(xml.id == "CSGAnimZPrecession");
	DeclinationDeg  = (float)atof(xml["DeclinationDeg"].c_str());
	AngularSpeedDPS = (float)atof(xml["AngularSpeedDPS"].c_str());
	StartAngleDeg   = (float)atof(xml["StartAngleDeg"].c_str());
}
