#pragma once
#include <list>
#include <vcg/xml/xml.h>
#include <vcg/Matrix44.h>
#include <vcg/xml/xmlUtility.h>

using namespace std;
namespace vcg {

class CSG;
class CSGNode 
{
public:
	virtual	~CSGNode(void){};
	
	virtual void glDraw(const float DocTime)=0;
	virtual void XMLWrite(FILE *fp)=0;
	virtual void XMLRead(Xml &xml, CSG *Base)=0;
};

class CSGGroup :public CSGNode 
{
	public:
	bool SaveLightState;
	CSGGroup() {SaveLightState=true;};
  virtual ~CSGGroup();
  virtual void Clear();
		typedef list<CSGNode *>::iterator iterator;
  
  
	list<CSGNode *> Sons;
	virtual void glDraw(const float DocTime);
  virtual void XMLWrite(FILE *fp);
  virtual void XMLRead(Xml &xml, CSG *Base);
};

class CSGTransformation :public CSGNode 
{
	public:
  Matrix44f m;
	virtual void glDraw(const float DocTime);
	virtual void XMLWrite(FILE *fp);
	virtual void XMLRead(Xml &xml, CSG *Base);
};

class CSGLight :public CSGNode 
{
	public:
	CSGLight(int id=0);
	int Id;
	Point4f Position;
	Point4f AmbientColor;
	Point4f DiffuseColor;
	Point4f SpecularColor;

	virtual void glDraw(const float DocTime);
	virtual void XMLWrite(FILE *fp);
	virtual void XMLRead(Xml &xml, CSG *Base);
};

class CSGTranslation :public CSGNode 
{
	public:
	CSGTranslation::CSGTranslation() :V(Point3f(0,0,0)){};
	CSGTranslation::CSGTranslation(Point3f &v): V(v) {};
  Point3f V;
	virtual void glDraw(const float DocTime) { glTranslate(V); };
	virtual void XMLWrite(FILE *fp);
	virtual void XMLRead(Xml &xml, CSG *Base);
};

class CSGRotation :public CSGNode 
{
	public:
  Point3f Axis;
	float AngleDeg; 
  virtual void glDraw(const float DocTime){ glRotatef(AngleDeg,Axis[0],Axis[1],Axis[2]); };
	virtual void XMLWrite(FILE *fp);
	virtual void XMLRead(Xml &xml, CSG *Base);
};


class CSGAnimRotation :public CSGNode 
{
	public:
  Point3f Axis;
	float AngularSpeedDPS; //Degree Per Sec;
  float StartAngleDeg; 
	virtual void glDraw(const float DocTime);
	virtual void XMLWrite(FILE *fp);
	virtual void XMLRead(Xml &xml, CSG *Base);
};

class CSGAnimZPrecession :public CSGNode 
{
	public:
  CSGAnimZPrecession();
 
  float DeclinationDeg;
	float AngularSpeedDPS; //Degree Per Sec;
  float StartAngleDeg; 
	virtual ~CSGAnimZPrecession(){};
	virtual void glDraw(const float DocTime);
  virtual void XMLWrite(FILE *fp);
	virtual void XMLRead(Xml &xml, CSG *Base);
};



// Minimal Scene graph
// Solo i gruppi definiscono uno scope di trasformazioni
// l'ordine dei figli  significativo.
class CSG
{
private:
	int LightCounter;
public:
	CSG(void);
	~CSG(void);
	CSGGroup root;
	virtual CSGNode *Allocate(const string &classname);
	virtual void Clear();
	int Pick(int x, int y, const float &DocTime);

};

} // end namespace